"""This is the ideal user flow
"""
import pytest

from deepeval.metrics import HallucinationMetric

from deepeval.test_case import LLMTestCase
from deepeval import assert_test


def generate_llm_output(query: str):
    expected_output = "Our customer success phone line is 1200-231-231."
    return expected_output


def test_llm_output():
    input = "What is the customer success phone line?"
    context = ["Our customer success phone line is 1200-231-231."]
    test_case = LLMTestCase(
        input=input, actual_output=generate_llm_output(input), context=context
    )
    assert_test(test_case, [HallucinationMetric(threshold=0.5)])


def test_llm_output_custom():
    actual_output = "Dogs and cats hate to walk around the beach."
    context = ["Dogs and cats love to walk around the beach."]
    test_case = LLMTestCase(
        input="Placerholder", actual_output=actual_output, context=context
    )
    with pytest.raises(AssertionError):
        assert_test(test_case, [HallucinationMetric(threshold=0.5)])


def test_0():
    query = "How does photosynthesis work?"
    output = "Photosynthesis is the process by which green plants and some other organisms use sunlight to synthesize foods with the help of chlorophyll pigment."
    expected_output = "Photosynthesis is the process by which green plants and some other organisms use sunlight to synthesize food with the help of chlorophyll pigment."
    context = ["Biology"]

    test_case = LLMTestCase(
        input=query,
        actual_output=output,
        expected_output=expected_output,
        context=context,
    )
    metric = HallucinationMetric()
    with pytest.raises(AssertionError):
        assert_test(test_case, metrics=[metric])



import pytest
import os
from rich import print
from typing import Optional, Any
from deepeval.constants import PYTEST_RUN_TEST_NAME
from deepeval.test_run import test_run_manager


def pytest_sessionstart(session: pytest.Session):
    test_run_manager.save_to_disk = True
    try:
        test_run_manager.create_test_run(
            deployment=session.config.getoption("--deployment"),
            file_name=session.config.getoption("file_or_dir")[0],
        )
    except:
        test_run_manager.create_test_run()


def pytest_addoption(parser):
    parser.addoption(
        "--deployment",
        action="store_true",
        default=False,
        help="Enable deployment mode",
    )


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_protocol(
    item: pytest.Item, nextitem: Optional[pytest.Item]
) -> Optional[Any]:
    os.environ[PYTEST_RUN_TEST_NAME] = item.nodeid.split("::")[-1]
    return None  # continue with the default protocol


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_sessionfinish(session: pytest.Session, exitstatus):
    print("Running teardown with pytest sessionfinish...")

    yield


def pytest_terminal_summary(terminalreporter, exitstatus, config):
    for report in terminalreporter.getreports("skipped"):
        if report.skipped:
            reason = report.longreprtext.split("\n")[-1]
            print(f"Test {report.nodeid} was skipped. Reason: {reason}")



from typing import Optional, List, Union
from pydantic import BaseModel, Field
import json
from langchain_core.language_models import BaseChatModel

from deepeval.utils import trimToJson
from deepeval.test_case import LLMTestCase
from deepeval.metrics import BaseMetric
from deepeval.models import GPTModel
from deepeval.templates import AnswerRelevancyTemplate
from deepeval.progress_context import metrics_progress_context


class AnswerRelvancyVerdict(BaseModel):
    verdict: str
    key_point: str = Field(default=None)


class AnswerRelevancyMetric(BaseMetric):
    def __init__(
        self,
        threshold: float = 0.5,
        model: Optional[Union[str, BaseChatModel]] = None,
        include_reason: bool = True,
    ):
        self.threshold = threshold
        self.model = GPTModel(model=model)
        self.evaluation_model = self.model.get_model_name()
        self.include_reason = include_reason
        self.n = 5

    def measure(self, test_case: LLMTestCase) -> float:
        if (
            test_case.input is None
            or test_case.actual_output is None
            or test_case.retrieval_context is None
        ):
            raise ValueError(
                "Input, actual output, or retrieval context cannot be None"
            )
        with metrics_progress_context(self.__name__, self.evaluation_model):
            self.key_points: List[str] = self._generate_key_points(
                test_case.actual_output, "\n".join(test_case.retrieval_context)
            )
            self.verdicts: List[
                AnswerRelvancyVerdict
            ] = self._generate_verdicts(test_case.input)

            answer_relevancy_score = self._generate_score()

            self.reason = self._generate_reason(
                test_case.input, test_case.actual_output, answer_relevancy_score
            )
            self.success = answer_relevancy_score >= self.threshold
            self.score = answer_relevancy_score
            return self.score

    def _generate_score(self):
        if len(self.verdicts) == 0:
            return 0

        relevant_count = 0
        for verdict in self.verdicts:
            if verdict.verdict.strip().lower() != "no":
                relevant_count += 1

        return relevant_count / len(self.verdicts)

    def _generate_reason(
        self, original_question: str, answer: str, score: float
    ) -> str:
        if self.include_reason is False:
            return None

        irrelevant_points = []
        for verdict in self.verdicts:
            if verdict.verdict.strip().lower() == "no":
                irrelevant_points.append(verdict.key_point)

        prompt = AnswerRelevancyTemplate.generate_reason(
            irrelevant_points=irrelevant_points,
            original_question=original_question,
            answer=answer,
            score=format(score, ".2f"),
        )

        res = self.model(prompt)
        return res.content

    def _generate_verdicts(
        self, original_question: str
    ) -> List[AnswerRelvancyVerdict]:
        prompt = AnswerRelevancyTemplate.generate_verdicts(
            original_question=original_question, key_points=self.key_points
        )

        res = self.model(prompt)
        json_output = trimToJson(res.content)
        data = json.loads(json_output)
        verdicts = [AnswerRelvancyVerdict(**item) for item in data["verdicts"]]

        if len(verdicts) != len(self.key_points):
            raise ValueError("Number of verdicts generated does not equal.")

        for i in range(len(verdicts)):
            verdicts[i].key_point = self.key_points[i]

        return verdicts

    def _generate_key_points(
        self, answer: str, retrieval_context: str
    ) -> List[str]:
        prompt = AnswerRelevancyTemplate.generate_key_points(
            answer=answer, retrieval_context=retrieval_context
        )

        res = self.model(prompt)
        json_output = trimToJson(res.content)
        data = json.loads(json_output)
        return data["key_points"]

    def is_successful(self) -> bool:
        self.success = self.score >= self.threshold
        return self.success

    @property
    def __name__(self):
        return "Answer Relevancy"



from pydantic import BaseModel, Field
from typing import Optional, List, Dict


class Golden(BaseModel):
    input: str
    actual_output: Optional[str] = Field(None, alias="actualOutput")
    expected_output: Optional[str] = Field(None, alias="expectedOutput")
    context: Optional[list] = Field(None)
    retrieval_context: Optional[list] = Field(None)
    additional_metadata: Optional[Dict] = Field(
        None, alias="additionalMetadata"
    )


class APIDataset(BaseModel):
    alias: str
    overwrite: bool
    goldens: Optional[List[Golden]] = Field(default=None)


class CreateDatasetHttpResponse(BaseModel):
    link: str


class DatasetHttpResponse(BaseModel):
    goldens: List[Golden]



from typing import Union, List, Dict

from transformers import (
    TrainerCallback,
    ProgressCallback,
    Trainer,
    TrainingArguments,
    TrainerState,
    TrainerControl,
)

from deepeval.metrics import BaseMetric
from deepeval.evaluate import execute_test
from deepeval.dataset import EvaluationDataset

from .utils import get_column_order, generate_test_cases
from .rich_manager import RichManager


class DeepEvalCallback(TrainerCallback):
    """
    Custom callback for deep evaluation during model training.

    Args:
        metrics (Union[BaseMetric, List[BaseMetric]]): Evaluation metrics.
        evaluation_dataset (EvaluationDataset): Dataset for evaluation.
        tokenizer_args (Dict): Arguments for the tokenizer.
        aggregation_method (str): Method for aggregating metric scores.
        trainer (Trainer): Model trainer.
    """

    def __init__(
        self,
        metrics: Union[BaseMetric, List[BaseMetric]] = None,
        evaluation_dataset: EvaluationDataset = None,
        tokenizer_args: Dict = None,
        aggregation_method: str = "avg",
        trainer: Trainer = None,
        show_table: bool = False,
        show_table_every: int = 1,
    ) -> None:
        super().__init__()

        self.show_table = show_table
        self.show_table_every = show_table_every
        self.metrics = metrics
        self.evaluation_dataset = evaluation_dataset
        self.tokenizer_args = tokenizer_args
        self.aggregation_method = aggregation_method
        self.trainer = trainer

        self.task_descriptions = {
            "generating": "[blue][STATUS] [white]Generating output from model (might take up few minutes)",
            "training": "[blue][STATUS] [white]Training in Progress",
            "evaluate": "[blue][STATUS] [white]Evaluating test-cases (might take up few minutes)",
            "training_end": "[blue][STATUS] [white]Training Ended",
        }

        self.train_bar_started = False
        self.epoch_counter = 0
        self.deepeval_metric_history = []

        total_train_epochs = self.trainer.args.num_train_epochs
        self.rich_manager = RichManager(show_table, total_train_epochs)
        self.trainer.remove_callback(ProgressCallback)

    def _calculate_metric_scores(self) -> Dict[str, List[float]]:
        """
        Calculate final evaluation scores based on metrics and test cases.

        Returns:
            Dict[str, List[float]]: Metric scores for each test case.
        """
        test_results = execute_test(
            test_cases=self.evaluation_dataset.test_cases, metrics=self.metrics
        )
        scores = {}
        for test in test_results:
            for metric in test.metrics:
                metric_name = str(metric.__name__).lower().replace(" ", "_")
                metric_score = metric.score
                scores.setdefault(metric_name, []).append(metric_score)

        scores = self._aggregate_scores(scores)
        return scores

    def _aggregate_scores(
        self, scores: Dict[str, List[float]]
    ) -> Dict[str, float]:
        """
        Aggregate metric scores using the specified method.

        Args:
            aggregation_method (str): Method for aggregating scores.
            scores (Dict[str, List[float]]): Metric scores for each test case.

        Returns:
            Dict[str, float]: Aggregated metric scores.
        """
        aggregation_functions = {
            "avg": lambda x: sum(x) / len(x),
            "max": max,
            "min": min,
        }
        if self.aggregation_method not in aggregation_functions:
            raise ValueError(
                "Incorrect 'aggregation_method', only accepts ['avg', 'min, 'max']"
            )
        return {
            key: aggregation_functions[self.aggregation_method](value)
            for key, value in scores.items()
        }

    def on_epoch_begin(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        """
        Event triggered at the begining of each training epoch.
        """
        self.epoch_counter += 1

    def on_epoch_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        """
        Event triggered at the end of each training epoch.
        """
        control.should_log = True
        self.rich_manager.change_spinner_text(
            self.task_descriptions["generating"]
        )
        test_cases = generate_test_cases(
            self.trainer.model,
            self.trainer.tokenizer,
            self.tokenizer_args,
            self.evaluation_dataset,
        )
        self.evaluation_dataset.test_cases = test_cases

    def on_log(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        """
        Event triggered after logging the last logs.
        """
        if (
            self.show_table
            and len(state.log_history) <= self.trainer.args.num_train_epochs
        ):
            self.rich_manager.advance_progress()

            if self.epoch_counter % self.show_table_every == 0:
                self.rich_manager.change_spinner_text(
                    self.task_descriptions["evaluate"]
                )

                scores = self._calculate_metric_scores()
                self.deepeval_metric_history.append(scores)
                self.deepeval_metric_history[-1].update(state.log_history[-1])

                self.rich_manager.change_spinner_text(
                    self.task_descriptions["training"]
                )
                columns = self._generate_table()
                self.rich_manager.update(columns)

    def _generate_table(self):
        """
        Generates table, along with progress bars

        Returns:
            rich.Columns: contains table and 2 progress bars
        """
        column, table = self.rich_manager.create_column()
        order = get_column_order(self.deepeval_metric_history[-1])

        if self.show_table:
            for key in order:
                table.add_column(key)

            for row in self.deepeval_metric_history:
                table.add_row(*[str(row[value]) for value in order])

        return column

    def on_train_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        """
        Event triggered at the end of model training.
        """
        self.rich_manager.change_spinner_text(
            self.task_descriptions["training_end"]
        )
        self.rich_manager.stop()

    def on_train_begin(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        """
        Event triggered at the begining of model training.
        """
        self.rich_manager.start()
        self.rich_manager.change_spinner_text(
            self.task_descriptions["training"]
        )



import pytest
from deepeval.test_case import LLMTestCase
from deepeval.metrics.ragas import (
    RagasMetric,
    RAGASContextualPrecisionMetric,
    RAGASContextualRelevancyMetric,
    RAGASFaithfulnessMetric,
    RAGASContextualRecallMetric,
    ConcisenessMetric,
    CorrectnessMetric,
    CoherenceMetric,
    MaliciousnessMetric,
    RAGASAnswerRelevancyMetric,
)
from deepeval import assert_test

query = "Who won the FIFA World Cup in 2018 and what was the score?"
output = "Winners of the FIFA world cup were the French national football team"
expected_output = "French national football team"
context = [
    "The FIFA World Cup in 2018 was won by the French national football team.",
    "I am birdy",
    "I am a froggy",
    "The French defeated Croatia 4-2 in the final FIFA match to claim the championship.",
]


@pytest.mark.skip(reason="openai is expensive")
def test_ragas_score():
    test_case = LLMTestCase(
        input=query,
        actual_output=output,
        expected_output=expected_output,
        context=context,
    )
    metric = RagasMetric()

    with pytest.raises(AssertionError):
        assert_test(
            test_case=[test_case],
            metrics=[metric],
        )


@pytest.mark.skip(reason="openai is expensive")
def test_everything():
    test_case = LLMTestCase(
        input=query,
        actual_output=output,
        expected_output=expected_output,
        retrieval_context=context,
        context=context,
    )
    metric1 = RAGASContextualRelevancyMetric(model="gpt-4")
    metric2 = RAGASFaithfulnessMetric(model="gpt-4")
    metric3 = RAGASContextualRecallMetric(model="gpt-4")
    metric4 = ConcisenessMetric(model="gpt-4")
    metric5 = CorrectnessMetric(model="gpt-4")
    metric6 = CoherenceMetric(model="gpt-4")
    metric7 = MaliciousnessMetric(model="gpt-4")
    metric8 = RAGASAnswerRelevancyMetric(model="gpt-4")
    metric9 = RAGASContextualPrecisionMetric()
    metric10 = RagasMetric()
    assert_test(
        test_case,
        [
            # metric1,
            # metric2,
            # metric3,
            # metric4,
            # metric5,
            # metric6,
            # metric7,
            # metric8,
            # metric9,
            metric10,
        ],
    )



from typing import List, Union

from transformers.trainer_callback import TrainerCallback

# from deepeval.experimental import BaseEvaluationExperiment


class DeepEvalHarnessCallback(TrainerCallback):
    """
    A [transformers.TrainerCallback] that logs various harness LLM evaluation metrics to DeepEval
    """

    def __init__(self, experiments):
        super().__init__()
        self.experiments = experiments

        raise NotImplementedError("DeepEvalHarnessCallback is WIP")



"""Metric for bias classifier - using the same min score subtraction methodology as the toxic classifier 
Rationale for bias classifier is described here https://arxiv.org/pdf/2208.05777.pdf
1 - Not Biased
0 - Bias
"""

from typing import Optional, List
from deepeval.metrics import BaseMetric
from deepeval.test_case import LLMTestCase, LLMTestCaseParams
from deepeval.scorer import Scorer


class BiasMetric(BaseMetric):
    def __init__(
        self,
        model_name: str = "original",
        threshold: float = 0.5,
    ):  # see paper for rationale https://arxiv.org/pdf/2208.05777.pdf
        self.model_name = model_name
        self.threshold = threshold

    def measure(self, test_case: LLMTestCase):
        if test_case.input is None or test_case.actual_output is None:
            raise ValueError("Input or actual output cannot be None")

        result = Scorer.neural_bias_score(
            test_case.actual_output, model=self.model_name
        )
        if result[0]["label"] == "Biased":
            bias_score = 0.5 + (result[0]["score"] / 2)
        else:
            bias_score = 0.5 - (result[0]["score"] / 2)

        self.success = bias_score <= self.threshold
        self.score = bias_score

        return self.score

    def is_successful(self) -> bool:
        self.success = self.score <= self.threshold
        return self.success

    @property
    def __name__(self):
        return "Bias"



from typing import List
from deepeval.dataset.api import Golden
from deepeval.test_case import LLMTestCase


def convert_test_cases_to_goldens(
    test_cases: List[LLMTestCase],
) -> List[Golden]:
    goldens = []
    for test_case in test_cases:
        golden = {
            "input": test_case.input,
            "actualOutput": test_case.actual_output,
            "expectedOutput": test_case.expected_output,
            "context": test_case.context,
        }
        goldens.append(Golden(**golden))
    return goldens


def convert_goldens_to_test_cases(goldens: List[Golden]) -> List[LLMTestCase]:
    test_cases = []
    for golden in goldens:
        test_case = LLMTestCase(
            input=golden.input,
            actual_output=golden.actual_output,
            expected_output=golden.expected_output,
            context=golden.context,
            retrieval_context=golden.retrieval_context,
        )
        test_cases.append(test_case)
    return test_cases



from dataclasses import dataclass
from typing import List, Optional
from enum import Enum


class LLMTestCaseParams(Enum):
    INPUT = "input"
    ACTUAL_OUTPUT = "actual_output"
    EXPECTED_OUTPUT = "expected_output"
    CONTEXT = "context"
    RETRIEVAL_CONTEXT = "retrieval_context"


@dataclass
class LLMTestCase:
    def __init__(
        self,
        input: str,
        actual_output: str,
        expected_output: Optional[str] = None,
        context: Optional[List[str]] = None,
        retrieval_context: Optional[List[str]] = None,
        latency: Optional[float] = None,
        cost: Optional[float] = None,
        id: Optional[str] = None,
    ):
        self.id = id
        self.input = input
        self.actual_output = actual_output
        self.expected_output = expected_output
        self.context = context
        self.retrieval_context = retrieval_context
        self.latency = latency
        self.cost = cost



from typing import List
from deepeval.test_case import LLMTestCase, LLMTestCaseParams
from deepeval.metrics.base_metric import BaseMetric
from deepeval.scorer import Scorer


class ToxicityMetric(BaseMetric):
    def __init__(
        self,
        model_name: str = "original",
        threshold: float = 0.5,
    ):
        self.threshold, self.model_name = threshold, model_name

    def measure(self, test_case: LLMTestCase):
        if test_case.input is None or test_case.actual_output is None:
            raise ValueError("Input or actual output cannot be None")
        _, results = Scorer.neural_toxic_score(
            prediction=test_case.actual_output, model=self.model_name
        )
        # sample output
        # {'toxicity': 0.98057544,
        # 'severe_toxicity': 0.106649496,
        # 'obscene': 0.32923067,
        # 'threat': 0.018646086,
        # 'insult': 0.7514423,
        # 'identity_attack': 0.86643445}
        toxicity_score = results["toxicity"]

        self.success = toxicity_score <= self.threshold
        self.score = toxicity_score
        return self.score

    def is_successful(self) -> bool:
        self.success = self.score <= self.threshold
        return self.success

    @property
    def __name__(self):
        return "Toxicity"



import pytest
from deepeval.metrics import JudgementalGPT
from deepeval.test_case import LLMTestCase, LLMTestCaseParams
from deepeval.types import Languages
from deepeval import assert_test


@pytest.mark.skip(reason="openai is expensive")
def test_judgemntal():
    test_case = LLMTestCase(
        input="I am a plant",
        actual_output="thanks for letting me know you're a plant",
    )

    metric = JudgementalGPT(
        name="Coherence",
        criteria="Coherence - determine whether 'actual output' follows from the 'input'.",
        evaluation_params=[
            LLMTestCaseParams.ACTUAL_OUTPUT,
            LLMTestCaseParams.INPUT,
        ],
        language=Languages.SPANISH,
        threshold=0.5,
    )

    assert_test(test_case, [metric])



from typing import List, Optional, Dict, Union
from threading import Thread, Lock
import json
from pydantic import BaseModel, Field
from langchain_core.language_models import BaseChatModel

from deepeval.test_case import LLMTestCase
from deepeval.metrics import BaseMetric
from deepeval.utils import trimToJson
from deepeval.models import GPTModel
from deepeval.templates import FaithfulnessTemplate
from deepeval.progress_context import metrics_progress_context


class FaithfulnessVerdict(BaseModel):
    verdict: str
    reason: str = Field(default=None)
    truth: str = Field(default=None)


class FaithfulnessMetric(BaseMetric):
    def __init__(
        self,
        threshold: float = 0.5,
        model: Optional[Union[str, BaseChatModel]] = None,
        include_reason: bool = True,
    ):
        self.threshold = threshold
        self.model = GPTModel(model=model)
        self.evaluation_model = self.model.get_model_name()
        self.include_reason = include_reason

    def measure(self, test_case: LLMTestCase):
        if (
            test_case.input is None
            or test_case.actual_output is None
            or test_case.retrieval_context is None
        ):
            raise ValueError(
                "Input, actual output, or retrieval context cannot be None"
            )
        with metrics_progress_context(self.__name__, self.evaluation_model):
            self.truths_list: List[List[str]] = self._generate_truths_list(
                test_case.retrieval_context
            )
            self.verdicts_list: List[
                List[FaithfulnessVerdict]
            ] = self._generate_verdicts_list(
                self.truths_list, test_case.actual_output
            )
            faithfulness_score = self._generate_score()
            self.reason = self._generate_reason(faithfulness_score)
            self.success = faithfulness_score >= self.threshold
            self.score = faithfulness_score
            return self.score

    def _generate_score(self):
        total_verdicts = 0
        faithful_count = 0
        for verdicts in self.verdicts_list:
            total_verdicts += len(verdicts)
            for verdict in verdicts:
                if verdict.verdict.strip().lower() != "no":
                    faithful_count += 1

        if total_verdicts == 0:
            return 0

        return faithful_count / total_verdicts

    def _generate_reason(self, score: float):
        if self.include_reason is False:
            return None

        contradictions = []
        for index, verdicts in enumerate(self.verdicts_list):
            for verdict in verdicts:
                if verdict.verdict.strip().lower() == "no":
                    data = {"contradiction": verdict.reason, "rank": index + 1}
                    contradictions.append(data)

        prompt: dict = FaithfulnessTemplate.generate_reason(
            contradictions=contradictions,
            score=format(score, ".2f"),
        )

        res = self.model(prompt)
        return res.content

    def _generate_truths(
        self,
        context: str,
        truths_list: List[str],
        lock: Lock,
    ):
        prompt = FaithfulnessTemplate.generate_truths(text=context)
        res = self.model(prompt)
        json_output = trimToJson(res.content)
        data = json.loads(json_output)
        truths = data["truths"]

        with lock:
            truths_list.append(truths)

    def _generate_truths_list(
        self, retrieval_context: List[str]
    ) -> List[List[str]]:
        truths_list: List[List[str]] = []
        threads = []
        lock = Lock()

        for context in retrieval_context:
            thread = Thread(
                target=self._generate_truths,
                args=(context, truths_list, lock),
            )
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

        return truths_list

    def _generate_verdicts(
        self,
        truths: List[str],
        text: str,
        verdicts_list: List[List[FaithfulnessVerdict]],
        lock: Lock,
    ):
        prompt = FaithfulnessTemplate.generate_verdicts(
            truths=truths, text=text
        )

        res = self.model(prompt)
        json_output = trimToJson(res.content)
        data = json.loads(json_output)
        verdicts = [FaithfulnessVerdict(**item) for item in data["verdicts"]]

        if len(verdicts) != len(truths):
            raise ValueError(
                "Number of verdicts generated does not equal truths."
            )

        for i in range(len(verdicts)):
            verdicts[i].truth = truths[i]

        with lock:
            verdicts_list.append(verdicts)

    def _generate_verdicts_list(
        self, truths_list: List[List[str]], text: str
    ) -> List[List[FaithfulnessVerdict]]:
        verdicts_list: List[List[FaithfulnessVerdict]] = []
        threads = []
        lock = Lock()

        for truths in truths_list:
            thread = Thread(
                target=self._generate_verdicts,
                args=(truths, text, verdicts_list, lock),
            )
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

        return verdicts_list

    def is_successful(self) -> bool:
        self.success = self.score >= self.threshold
        return self.success

    @property
    def __name__(self):
        return "Faithfulness"



import json
from typing import Optional, List, Tuple, Union
from langchain_core.language_models import BaseChatModel

from deepeval.metrics import BaseMetric
from deepeval.test_case import LLMTestCase, LLMTestCaseParams
from deepeval.templates import (
    evaluation_steps_template,
    evaluation_results_template,
)
from deepeval.utils import trimToJson
from deepeval.models import GPTModel

from pydantic import BaseModel


class GEvalResponse(BaseModel):
    score: float
    reason: str


class GEval(BaseMetric):
    def __init__(
        self,
        name: str,
        evaluation_params: List[LLMTestCaseParams],
        criteria: Optional[str] = None,
        evaluation_steps: Optional[List[str]] = None,
        model: Optional[Union[str, BaseChatModel]] = None,
        threshold: float = 0.5,
    ):
        self.name = name
        self.evaluation_params = evaluation_params

        # Check if both criteria and evaluation_steps are not None at the same time
        if criteria is None and evaluation_steps is None:
            raise ValueError(
                "Either 'criteria' or 'evaluation_steps' must be provided, but not both None."
            )

        # Check if criteria is provided, it cannot be an empty string
        if criteria is not None and not criteria.strip():
            raise ValueError("Criteria provided cannot be an empty string.")

        # Check if evaluation_steps is provided, it cannot be an empty list
        if evaluation_steps is not None and len(evaluation_steps) == 0:
            raise ValueError(
                "Evaluation steps must not be an empty list. Either omit evaluation steps or include a non-empty list of steps."
            )

        self.criteria = criteria
        self.model = GPTModel(model=model)
        self.evaluation_model = self.model.get_model_name()
        self.evaluation_steps = evaluation_steps
        self.threshold = threshold

    def measure(self, test_case: LLMTestCase):
        """LLM evaluated metric based on the GEval framework: https://arxiv.org/pdf/2303.16634.pdf"""

        # Measure the test case
        for param in self.evaluation_params:
            if (
                not hasattr(test_case, param.value)
                or getattr(test_case, param.value) is None
            ):
                raise ValueError(
                    f"Test case is missing the required attribute: {param.value}"
                )

        if self.evaluation_steps is None:
            json_output = trimToJson(self.generate_evaluation_steps())
            data = json.loads(json_output)
            self.evaluation_steps = data["steps"]

        score, reason = self.evaluate(test_case)
        self.reason = reason
        self.score = float(score) / 10
        self.success = score >= self.threshold
        return self.score

    def is_successful(self) -> bool:
        self.success = self.score >= self.threshold
        return self.success

    def generate_evaluation_steps(self):
        prompt: dict = evaluation_steps_template.format(criteria=self.criteria)

        res = self.model(prompt)

        return res.content

    def evaluate(self, test_case: LLMTestCase) -> Tuple[int, str]:
        text = """"""

        for param in self.evaluation_params:
            value = getattr(test_case, param.value)
            text += f"{param.value}: {value} \n\n"

        prompt: dict = evaluation_results_template.format(
            evaluation_steps=self.numbered_evaluation_steps(),
            text=text,
        )

        res = self.model(prompt)
        json_output = trimToJson(res.content)
        data = json.loads(json_output)

        return data["score"], data["reason"]

    def numbered_evaluation_steps(self):
        evaluation_steps = """"""
        for index, string in enumerate(self.evaluation_steps, start=1):
            evaluation_steps += f"{index}. {string}\n"

        return evaluation_steps

    @property
    def __name__(self):
        return self.name



import typer
from typing_extensions import Annotated

# Rich has a few dependency issues
try:
    from rich import print
except Exception as e:
    pass
from deepeval.key_handler import KEY_FILE_HANDLER, KeyValues
from deepeval.cli.test import app as test_app
from typing import Optional
import webbrowser

app = typer.Typer(name="deepeval")

app.add_typer(test_app, name="test")


@app.command()
def login(
    api_key: Annotated[
        str,
        typer.Option(
            help="API key to get from https://app.confident-ai.com. Required if you want to log events to the server."
        ),
    ] = "",
    confident_api_key: Optional[str] = typer.Option(
        None,
        "--confident-api-key",
        "-c",
        help="Optional confident API key to bypass login.",
    ),
):
    # Use the confident_api_key if it is provided, otherwise proceed with existing logic
    if confident_api_key:
        api_key = confident_api_key
    else:
        """Login to the DeepEval platform."""
        print("Welcome to [bold]DeepEval[/bold]!")
        print(
            "Login and grab your API key here: [link=https://app.confident-ai.com]https://app.confident-ai.com[/link] "
        )
        webbrowser.open("https://app.confident-ai.com")
        if api_key == "":
            while True:
                api_key = input("Paste your API Key: ").strip()
                if api_key:
                    break
                else:
                    print("API Key cannot be empty. Please try again.\n")

    KEY_FILE_HANDLER.write_key(KeyValues.API_KEY, api_key)
    print("Congratulations! Login successful :raising_hands: ")
    print(
        "If you are new to DeepEval, follow our quickstart tutorial here: [bold][link=https://docs.confident-ai.com/docs/getting-started]https://docs.confident-ai.com/docs/getting-started[/link][/bold]"
    )


@app.command(name="set-azure-openai")
def set_azure_openai_env(
    azure_openai_api_key: str = typer.Option(
        ..., "--openai-api-key", help="Azure OpenAI API key"
    ),
    azure_openai_endpoint: str = typer.Option(
        ..., "--openai-endpoint", help="Azure OpenAI endpoint"
    ),
    openai_api_version: str = typer.Option(
        ..., "--openai-api-version", help="OpenAI API version"
    ),
    azure_deployment_name: str = typer.Option(
        ..., "--deployment-name", help="Azure deployment name"
    ),
    azure_model_version: Optional[str] = typer.Option(
        None, "--model-version", help="Azure model version (optional)"
    ),
):
    KEY_FILE_HANDLER.write_key(
        KeyValues.AZURE_OPENAI_API_KEY, azure_openai_api_key
    )
    KEY_FILE_HANDLER.write_key(
        KeyValues.AZURE_OPENAI_ENDPOINT, azure_openai_endpoint
    )
    KEY_FILE_HANDLER.write_key(KeyValues.OPENAI_API_VERSION, openai_api_version)
    KEY_FILE_HANDLER.write_key(
        KeyValues.AZURE_DEPLOYMENT_NAME, azure_deployment_name
    )

    if azure_model_version is not None:
        KEY_FILE_HANDLER.write_key(
            KeyValues.AZURE_MODEL_VERSION, azure_model_version
        )

    KEY_FILE_HANDLER.write_key(KeyValues.USE_AZURE_OPENAI, "YES")

    print(
        ":raising_hands: Congratulations! You're now using Azure OpenAI for all evals that require an LLM."
    )


@app.command(name="unset-azure-openai")
def unset_azure_openai_env():
    KEY_FILE_HANDLER.remove_key(KeyValues.AZURE_OPENAI_API_KEY)
    KEY_FILE_HANDLER.remove_key(KeyValues.AZURE_OPENAI_ENDPOINT)
    KEY_FILE_HANDLER.remove_key(KeyValues.OPENAI_API_VERSION)
    KEY_FILE_HANDLER.remove_key(KeyValues.AZURE_DEPLOYMENT_NAME)
    KEY_FILE_HANDLER.remove_key(KeyValues.AZURE_MODEL_VERSION)
    KEY_FILE_HANDLER.remove_key(KeyValues.USE_AZURE_OPENAI)

    print(
        ":raising_hands: Congratulations! You're now using regular OpenAI for all evals that require an LLM."
    )


if __name__ == "__main__":
    app()



"""Test for callbacks
"""

from transformers import (
    Seq2SeqTrainer,
    Seq2SeqTrainingArguments,
    T5Tokenizer,
    T5ForConditionalGeneration,
    DataCollatorForSeq2Seq,
)

from datasets import load_dataset

from deepeval.callbacks.huggingface import DeepEvalCallback
from deepeval.metrics import HallucinationMetric, AnswerRelevancyMetric
from deepeval.dataset import EvaluationDataset, Golden

import os
import random

os.environ["TRANSFORMERS_NO_ADVISORY_WARNINGS"] = "true"
os.environ["OPENAI_API_KEY"] = "API-KEY"


def create_prompt(row):
    """Merge Context and Question into a single string"""
    contexts = row["context"]["contexts"]
    question = row["question"]
    prompt = f"""{'CONTEXT: ' + str("; ".join(contexts)) if contexts else ''}
            QUESTION: {question}
            ANSWER:"""
    return {"input": prompt, "response": row["long_answer"]}


def prepare_dataset(tokenizer, tokenizer_args):
    dataset = load_dataset("pubmed_qa", "pqa_labeled")
    merged_dataset = dataset.map(
        create_prompt,
        remove_columns=[
            "question",
            "context",
            "long_answer",
            "pubid",
            "final_decision",
        ],
    )

    def tokenize_text(dataset, padding="max_length"):
        model_input = tokenizer(dataset["input"], **tokenizer_args)
        response = tokenizer(dataset["response"], **tokenizer_args)

        if padding == "max_length":
            response["input_ids"] = [
                [(l if l != tokenizer.pad_token_id else -100) for l in label]
                for label in response["input_ids"]
            ]

        model_input["labels"] = response["input_ids"]
        return model_input

    tokenized_dataset = merged_dataset.map(
        tokenize_text, remove_columns=["input", "response"]
    )
    tokenized_dataset = tokenized_dataset.map(
        lambda x: {
            "input_ids": x["input_ids"][0],
            "labels": x["labels"][0],
            "attention_mask": x["attention_mask"][0],
        }
    )
    return dataset, merged_dataset, tokenized_dataset


def create_deepeval_dataset(dataset, sample_size):
    total_length = len(dataset)
    random_index_list = [
        random.randint(0, total_length) for _ in range(sample_size)
    ]
    print(random_index_list)
    eval_dataset = [dataset[row] for row in random_index_list]
    goldens = []
    for row in eval_dataset:
        context = ["; ".join(row["context"]["contexts"])]
        golden = Golden(
            input=row["question"],
            expectedOutput=row["long_answer"],
            context=context,
            retrieval_context=context,
        )
        goldens.append(golden)

    return EvaluationDataset(goldens=goldens)


if __name__ == "__main__":
    # initialize tokenizer
    tokenizer = T5Tokenizer.from_pretrained("google/flan-t5-small")

    # initalize model
    model = T5ForConditionalGeneration.from_pretrained("google/flan-t5-small")
    model.resize_token_embeddings(len(tokenizer))

    # create tokenized dataset
    tokenizer_args = {
        "return_tensors": "pt",
        "max_length": 128,
        "padding": "max_length",
        "truncation": True,
        "padding": True,
    }

    dataset, merged_dataset, tokenized_dataset = prepare_dataset(
        tokenizer, tokenizer_args
    )

    label_pad_token_id = -100
    data_collator = DataCollatorForSeq2Seq(
        tokenizer,
        model=model,
        label_pad_token_id=label_pad_token_id,
        pad_to_multiple_of=8,
    )

    repository_id = f"flan-t5-small"

    # Define training args
    training_args = Seq2SeqTrainingArguments(
        output_dir=repository_id,
        overwrite_output_dir=True,
        num_train_epochs=50,
        per_device_train_batch_size=8,
    )

    # Create Trainer instance
    trainer = Seq2SeqTrainer(
        model=model,
        tokenizer=tokenizer,
        args=training_args,
        data_collator=data_collator,
        train_dataset=tokenized_dataset["train"],
    )

    eval_dataset = create_deepeval_dataset(dataset["train"], sample_size=5)
    hallucination_metric = HallucinationMetric(threshold=0.3)
    answer_relevancy_metric = AnswerRelevancyMetric(
        threshold=0.5, model="gpt-3.5-turbo"
    )
    metrics = [hallucination_metric, answer_relevancy_metric]

    # initalize DeepEvalCallback
    callback = DeepEvalCallback(
        metrics=metrics,
        evaluation_dataset=eval_dataset,
        tokenizer_args=tokenizer_args,
        trainer=trainer,
        show_table=True,
        show_table_every=1,
    )
    trainer.add_callback(callback)
    trainer.train()



from enum import Enum
import copy
import os
import time
from typing import Any
from collections.abc import Iterable
import tqdm
import re
import string
import numpy as np
from dataclasses import asdict, is_dataclass
import re

from deepeval.key_handler import KeyValues, KEY_FILE_HANDLER


def is_confident():
    confident_api_key = KEY_FILE_HANDLER.fetch_data(KeyValues.API_KEY)
    return confident_api_key is not None


def drop_and_copy(obj, drop_attrs):
    # Function to drop attributes from a single object
    def drop_attrs_from_single_obj(single_obj, drop_attrs):
        temp_attrs = {}
        for attr in drop_attrs:
            if hasattr(single_obj, attr):
                temp_attrs[attr] = getattr(single_obj, attr)
                delattr(single_obj, attr)
        return temp_attrs

    # Check if obj is iterable (but not a string)
    if isinstance(obj, Iterable) and not isinstance(obj, str):
        copied_objs = []
        for item in obj:
            temp_attrs = drop_attrs_from_single_obj(item, drop_attrs)
            copied_objs.append(copy.deepcopy(item))
            # Restore attributes to the original item
            for attr, value in temp_attrs.items():
                setattr(item, attr, value)
        return copied_objs
    else:
        # If obj is not iterable, apply directly
        temp_attrs = drop_attrs_from_single_obj(obj, drop_attrs)
        copied_obj = copy.deepcopy(obj)
        # Restore attributes to the original object
        for attr, value in temp_attrs.items():
            setattr(obj, attr, value)
        return copied_obj


def dataclass_to_dict(instance: Any) -> Any:
    if is_dataclass(instance):
        return {k: dataclass_to_dict(v) for k, v in asdict(instance).items()}
    elif isinstance(instance, Enum):
        return instance.value
    elif isinstance(instance, list):
        return [dataclass_to_dict(item) for item in instance]
    elif isinstance(instance, tuple):
        return tuple(dataclass_to_dict(item) for item in instance)
    elif isinstance(instance, dict):
        return {k: dataclass_to_dict(v) for k, v in instance.items()}
    else:
        return instance


def trimToJson(input_string: str) -> str:
    start = input_string.find("{")
    end = input_string.rfind("}") + 1
    return input_string[start:end] if start != -1 and end != 0 else ""


def delete_file_if_exists(file_path):
    try:
        if os.path.exists(file_path):
            os.remove(file_path)
    except Exception as e:
        print(f"An error occurred: {e}")


def softmax(x):
    e_x = np.exp(x - np.max(x, axis=1, keepdims=True))
    return e_x / e_x.sum(axis=1, keepdims=True)


def cosine_similarity(vector_a, vector_b):
    dot_product = np.dot(vector_a, vector_b)
    norm_a = np.linalg.norm(vector_a)
    norm_b = np.linalg.norm(vector_b)
    similarity = dot_product / (norm_a * norm_b)
    return similarity


def chunk_text(text, chunk_size=20):
    words = text.split()
    chunks = [
        " ".join(words[i : i + chunk_size])
        for i in range(0, len(words), chunk_size)
    ]
    return chunks


def normalize_text(text: str) -> str:
    """Lower text and remove punctuation, articles and extra whitespace.
    Copied from the [QuAC](http://quac.ai/) evaluation script found at
    https://s3.amazonaws.com/my89public/quac/scorer.py"""

    def remove_articles(text: str) -> str:
        return re.sub(r"\b(a|an|the)\b", " ", text)

    def white_space_fix(text: str) -> str:
        return " ".join(text.split())

    def remove_punc(text: str) -> str:
        exclude = set(string.punctuation)
        return "".join(ch for ch in text if ch not in exclude)

    def lower(text: str) -> str:
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(text))))


###############################################
# Source: https://github.com/tingofurro/summac
###############################################

# GPU-related business


def get_freer_gpu():
    os.system("nvidia-smi -q -d Memory |grep -A4 GPU|grep Free >tmp_smi")
    memory_available = [
        int(x.split()[2]) + 5 * i
        for i, x in enumerate(open("tmp_smi", "r").readlines())
    ]
    os.remove("tmp_smi")
    return np.argmax(memory_available)


def any_gpu_with_space(gb_needed):
    os.system("nvidia-smi -q -d Memory |grep -A4 GPU|grep Free >tmp_smi")
    memory_available = [
        float(x.split()[2]) / 1024.0
        for i, x in enumerate(open("tmp_smi", "r").readlines())
    ]
    os.remove("tmp_smi")
    return any([mem >= gb_needed for mem in memory_available])


def wait_free_gpu(gb_needed):
    while not any_gpu_with_space(gb_needed):
        time.sleep(30)


def select_freer_gpu():
    freer_gpu = str(get_freer_gpu())
    print("Will use GPU: %s" % (freer_gpu))
    os.environ["CUDA_LAUNCH_BLOCKING"] = "1"
    os.environ["CUDA_VISIBLE_DEVICES"] = "" + freer_gpu
    return freer_gpu


def batcher(iterator, batch_size=4, progress=False):
    if progress:
        iterator = tqdm.tqdm(iterator)

    batch = []
    for elem in iterator:
        batch.append(elem)
        if len(batch) == batch_size:
            final_batch = batch
            batch = []
            yield final_batch
    if len(batch) > 0:  # Leftovers
        yield batch
